<?php
/*
 * Inventory history functions
 *
 * NOTICE OF LICENSE
 * This source file is subject to the Open Software License (OSL 3.0)
 * that is bundled with this package in the file LICENSE.TXT.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/OSL-3.0
 *
 * DISCLAIMER
 * Do not edit or add to this file if you wish to upgrade Bizuno to newer
 * versions in the future. If you wish to customize Bizuno for your
 * needs please refer to http://www.phreesoft.com for more information.
 *
 * @name       Bizuno ERP
 * @author     Dave Premo, PhreeSoft <support@phreesoft.com>
 * @copyright  2008-2023, PhreeSoft, Inc.
 * @license    http://opensource.org/licenses/OSL-3.0 Open Software License (OSL 3.0)
 * @version    6.x Last Update: 2023-04-18
 * @filesource /controllers/inventory/history.php
 */

namespace bizuno;

class inventoryHistory
{
    public $moduleID = 'inventory';

    function __construct()
    {
        $this->lang          = getLang($this->moduleID);
        $this->percent_diff  = 0.10; // the percentage differnece from current value to notify for adjustment
        $this->months_of_data= 12;   // valid values are 1, 3, 6, or 12
        $this->med_avg_diff  = 0.25; // the maximum percentage difference from the median and average, for large swings
    }

    /**
     * Ajax call to refresh the movement tab of an inventory item being edited.
     * @param array $layout - structure coming in
     * @return modified structure
     */
    public function movement(&$layout=[])
    {
        if (!$security = validateSecurity($this->moduleID, 'inv_mgr', 1)) { return; }
        $rID = clean('rID', 'integer', 'get');
        if (!$rID) { return ("This SKU does not have any history!"); }
        $sku   = dbGetValue(BIZUNO_DB_PREFIX.'inventory', 'sku', "id=$rID");
        $data = ['type'=>'divHTML',
            'divs'    => [
                'main' => ['order'=>50,'type'=>'divs','classes'=>['areaView'],'divs'=>[
                    'dgJ10' => ['order'=>20,'type'=>'panel','key'=>'dgJ10','classes'=>['block99']],
                    'dgJ04' => ['order'=>25,'type'=>'panel','key'=>'dgJ04','classes'=>['block99']],
                    'dgWO'  => ['order'=>30,'type'=>'panel','key'=>'dgWO', 'classes'=>['block99']], // forces it on it's own line so the rest line up
                    'dgJ09' => ['order'=>35,'type'=>'panel','key'=>'dgJ09','classes'=>['block99']],
                    'dgJ03' => ['order'=>40,'type'=>'panel','key'=>'dgJ03','classes'=>['block99']],
                    'dgJ12' => ['order'=>45,'type'=>'panel','key'=>'dgJ12','classes'=>['block99']],
                    'dgJ06' => ['order'=>50,'type'=>'panel','key'=>'dgJ06','classes'=>['block99']]]]],
            'panels'  => [
                'dgJ03' => ['type'=>'datagrid','key'=>'dgJ03'],
                'dgJ04' => ['type'=>'datagrid','key'=>'dgJ04'],
                'dgJ06' => ['type'=>'datagrid','key'=>'dgJ06'],
                'dgJ09' => ['type'=>'datagrid','key'=>'dgJ09'],
                'dgJ10' => ['type'=>'datagrid','key'=>'dgJ10'],
                'dgJ12' => ['type'=>'datagrid','key'=>'dgJ12'],
                'dgWO'  => ['type'=>'datagrid','key'=>'dgWO']],
            'datagrid'=> [
                'dgJ03' => $this->dgJ04J10( 3, $rID),
                'dgJ04' => $this->dgJ04J10( 4, $rID),
                'dgJ06' => $this->dgJ06J12( 6, $rID, $sku),
                'dgJ09' => $this->dgJ04J10( 9, $rID),
                'dgJ10' => $this->dgJ04J10(10, $rID),
                'dgJ12' => $this->dgJ06J12(12, $rID, $sku),
                'dgWO'  => $this->dgWO($rID)]];
        $layout = array_replace_recursive($layout, $data);
    }

    /**
     * Rows for the inventory movement grid for purchases and sales
     * @param type $layout
     */
    public function movementRows(&$layout=[])
    {
        $jID   = clean('jID',   'integer', 'get');
        $skuID = clean('skuID', 'integer', 'get');
        if (!$security = validateSecurity('phreebooks', "j{$jID}_mgr", 1)) { return; }
        $sku   = dbGetValue(BIZUNO_DB_PREFIX.'inventory', 'sku', "id=$skuID");
        $struc = $this->dgJ06J12($jID, $skuID, $sku);
        $layout= array_replace_recursive($layout, ['type'=>'datagrid','key'=>'manager','datagrid'=>['manager'=>$struc]]);
        msgDebug("\nLeaving movementRows with ;ayout = ".print_r($layout, true));
    }

    /**
     * Builds the rows for Work Order movement
     * @param array $layout
     * @return modified $layout
     */
    public function buildRows(&$layout=[])
    {
        if (!$security = validateSecurity('proInv', 'srvBldJrnl', 1)) { return; }
        $output = $skus = [];
        $skuID  = clean('skuID', 'integer', 'get');
        $sku    = dbGetValue(BIZUNO_DB_PREFIX.'inventory', 'sku', "id=$skuID");
        // get all open work orders, assume they are all assemblies
        $openWOs= dbGetMulti(BIZUNO_DB_PREFIX.'srvBuilder_journal', "closed='0'", '', ['id','sku_id','qty','sb_ref','store_id','due_date','create_date','notes','steps']);
        // search within the BOM for each skuID for this SKU
        foreach ($openWOs as $row) { $skus[] = $row['sku_id']; }
        $hits   = !empty($skus) ? dbGetMulti(BIZUNO_DB_PREFIX.'inventory_assy_list', "sku='".addslashes($sku)."' AND ref_id IN (".implode(',', $skus).")", '', ['ref_id', 'qty']) : [];
        $matches= [$skuID];
        $totalQ = [1];
        foreach ($hits as $hit) {
            $matches[] = $hit['ref_id'];
            $totalQ[]  = $hit['qty'];
        }
        msgDebug("\nMatches = ".print_r($matches, true));
        // build the output
        foreach ($openWOs as $wo) {
            $key = array_search($wo['sku_id'], $matches);
            if ($key===false) { continue; }
            // figure out if work order has been started
            $steps   = json_decode($wo['steps'], true);
            $started = !empty($steps[1]['complete']) ? 1 : 0;
            $output[]= [
                'id'         => $wo['id'],
                'started'    => $started,
                'sb_ref'     => $wo['sb_ref'],
                'store_id'   => viewFormat($wo['store_id'], 'contactID'),
                'create_date'=> viewFormat($wo['create_date'], 'date'),
                'due_date'   => viewFormat($wo['due_date'], 'date'),
                'qty'        => intval($wo['qty'] * $totalQ[$key]),
                'notes'      => substr($wo['notes'], 0, 30)];
        }
        $gridData = viewGridFilter($output);
        msgDebug("\nOutput = ".print_r($gridData, true));
        $layout = array_replace_recursive($layout, ['type'=>'raw','content'=>json_encode(['total'=>sizeof($output),'rows'=>$gridData])]);
    }

    /**
     * Generates lists of historical values for an inventory item for the past 12 months
     * @param array $layout - structure coming in
     * @return modified structure
     */
    public function historian(&$layout=[])
    {
        if (!$security = validateSecurity($this->moduleID, 'inv_mgr', 1)) { return; }
        $rID    = clean('rID', 'integer', 'get');
        if (!$rID) { return ("This SKU does not have any history!"); }
        $skuInfo= dbGetRow(BIZUNO_DB_PREFIX.'inventory', "id='$rID'");
        $history= $this->historyData($skuInfo);
        $usage  = "<div><br /><hr /><h2>".$this->lang['stock_usage_lbl'].'</h2>
<table style="width:100%"><thead class="panel-header"><tr><th style="width:50%">&nbsp;</th><th style="width:25%">'.lang('journal_main_journal_id_6').'</th><th style="width:25%">'.lang('journal_main_journal_id_12')."</th></tr></thead><tbody>
    <tr><td>".$this->lang['01month'].'</td><td style="text-align:center;">'.$history['01purch'].'</td><td style="text-align:center;">'.$history['01sales']."</td></tr>
    <tr><td>".$this->lang['03month'].'</td><td style="text-align:center;">'.$history['03purch'].'</td><td style="text-align:center;">'.$history['03sales']."</td></tr>
    <tr><td>".$this->lang['06month'].'</td><td style="text-align:center;">'.$history['06purch'].'</td><td style="text-align:center;">'.$history['06sales']."</td></tr>
    <tr><td>".$this->lang['12month'].'</td><td style="text-align:center;">'.$history['12purch'].'</td><td style="text-align:center;">'.$history['12sales'].'</td></tr>
</tbody></table>';
        if (!empty($history['id'])) { $usage .= '<br />'.lang('id').': '.$history['id']."\n"; }
        $usage  .= '</div>';
        $fields = [
            'create' => ['order'=>10,'label'=>lang('inventory_creation_date'),    'attr'=>['type'=>'date','readonly'=>true,'value'=>$skuInfo['creation_date']]],
            'update' => ['order'=>20,'label'=>lang('last_update'),                'attr'=>['type'=>'date','readonly'=>true,'value'=>$skuInfo['last_update']]],
            'journal'=> ['order'=>30,'label'=>lang('inventory_last_journal_date'),'attr'=>['type'=>'date','readonly'=>true,'value'=>$skuInfo['last_journal_date']]],
            'usage'  => ['order'=>60,'html'=>$usage, 'attr'=>['type'=>'raw']]];
        $data   = ['type'=>'divHTML',
            'divs'    =>[
                'main' => ['order'=>50,'type'=>'divs','classes'=>['areaView'],'divs'=>[
                    'usage'=> ['order'=>10,'type'=>'panel','key'=>'usage','classes'=>['block50']],
                    'dates'=> ['order'=>20,'type'=>'panel','key'=>'dates','classes'=>['block50']],
                    'dgJ06'=> ['order'=>50,'type'=>'panel','key'=>'dgJ06','classes'=>['block50']],
                    'dgJ12'=> ['order'=>60,'type'=>'panel','key'=>'dgJ12','classes'=>['block50']],
                    ]]],
            'panels'  => [
                'usage'=> ['label'=>lang('usage'),  'type'=>'fields','keys'=>['usage']],
                'dates'=> ['label'=>lang('details'),'type'=>'fields','keys'=>['create','update','journal']],
                'dgJ06'=> ['type'=>'datagrid','key'=>'dgJ06'],
                'dgJ12'=> ['type'=>'datagrid','key'=>'dgJ12']],
            'datagrid'=> [
                'dgJ06'=>$this->dgJ06J12Avg(6),
                'dgJ12'=>$this->dgJ06J12Avg(12)],
            'fields'  => $fields,
            'jsHead'  => ['init' =>"var dataJ6 = " .json_encode($history['purchases']).";\nvar dataJ12 = ".json_encode($history['sales']).";"]];
        msgDebug("\nReturning from inventory history with array = ".print_r($history, true));
        $layout = array_replace_recursive($layout, $data);
    }

    /**
     * Retrieves the historical PO/SO data for a given SKU
     * @param array $layout - structure coming in
     * @return modified $layout
     */
    public function historyRows(&$layout=[])
    {
        if (!$security = validateSecurity($this->moduleID, 'inv_mgr', 1)) { return; }
        $jID  = clean('jID',  'integer', 'get');
        $skuID= clean('skuID','integer', 'get');
        $sort = clean('sort', ['format'=>'cmd','default'=>'date_1'], 'post');
        $order= strtoupper(clean('order', ['format'=>'cmd','default'=>'ASC'], 'post'));
        $tail = "ORDER BY $sort $order";
        $data = [];
        $sku  = dbGetValue(BIZUNO_DB_PREFIX.'inventory', 'sku', "id=$skuID");
        $stmt = dbGetResult("SELECT m.id, m.journal_id, m.store_id, m.invoice_num, m.purch_order_id, m.primary_name_b, m.rep_id, m.waiting, i.qty, i.post_date, i.date_1,
            i.id AS item_id FROM ".BIZUNO_DB_PREFIX."journal_main m JOIN ".BIZUNO_DB_PREFIX."journal_item i ON m.id=i.ref_id
            WHERE m.journal_id='$jID' AND i.sku='$sku' AND m.closed='0' $tail");
        if ($stmt) {
            $result= $stmt->fetchAll(\PDO::FETCH_ASSOC);
            msgDebug("\nReturned number of open SO/PO rows = ".sizeof($result));
            foreach ($result as $row) {
                $adj = dbGetValue(BIZUNO_DB_PREFIX.'journal_item', "SUM(qty) AS qty", "gl_type='itm' AND item_ref_id={$row['item_id']}", false);
                if ($row['qty'] > $adj) { $data[] = [
                    'id'            => $row['id'],
                    'store_id'      => viewFormat($row['store_id'],'storeID'),
                    'rep_id'        => viewFormat($row['rep_id'],  'contactID'),
                    'waiting'       => $row['waiting'],
                    'invoice_num'   => $row['invoice_num'],
                    'primary_name_b'=> $row['primary_name_b'],
                    'purch_order_id'=> $row['purch_order_id'],
                    'post_date'     => viewDate($row['post_date']),
                    'qty'           => $row['qty'] - $adj,
                    'date_1'        => viewDate($row['date_1'])];
                }
            }
        }
        $layout = array_replace_recursive($layout, ['content'=>['total'=>sizeof($data),'rows'=>$data]]);
    }

    /**
     * Builds the history for a given sku
     * @param array $skuInfo
     * @return array
     */
    public function historyData($skuInfo=[], $verbose=true)
    {
        $history = ['id'=>$skuInfo['id'],'purchases'=>[],'sales'=>[]];
        // load the units received and sold, assembled and adjusted
        $dates = localeGetDates();
        $cur_month = $dates['ThisYear'].'-'.substr('0'.$dates['ThisMonth'], -2).'-01';
        for ($i = 0; $i < 13; $i++) {
            $index = substr($cur_month, 0, 7);
            $month = substr($index, 5, 2);
            $year  = substr($index, 0, 4);
            $history['purchases'][$index]= ['year'=>$year, 'month'=>lang('month_'.$month), 'qty'=>0, 'total'=>0, 'usage'=>0];
            $history['sales'][$index]    = ['year'=>$year, 'month'=>lang('month_'.$month), 'qty'=>0, 'total'=>0, 'usage'=>0];
            $cur_month = localeCalculateDate($cur_month, 0, -1, 0);
        }
        $next_month = localeCalculateDate($dates['ThisYear'].'-'.substr('0'.$dates['ThisMonth'], -2).'-01', 0, 1, 0);
        $last_year = ($dates['ThisYear']-1).'-'.substr('0'.$dates['ThisMonth'], -2).'-01';
        $sql = "SELECT m.journal_id, m.post_date, i.qty, i.gl_type, i.credit_amount, i.debit_amount
            FROM ".BIZUNO_DB_PREFIX."journal_main m JOIN ".BIZUNO_DB_PREFIX."journal_item i ON m.id=i.ref_id
            WHERE m.journal_id IN (6,7,12,13,14,16,19,21) AND i.sku='".addslashes($skuInfo['sku'])."' AND m.post_date>='$last_year' AND m.post_date<'$next_month'
            ORDER BY m.post_date DESC";
        $stmt  = dbGetResult($sql);
        $result2= $stmt->fetchAll(\PDO::FETCH_ASSOC);
        msgDebug("\nReturned monthly sales/purchases rows = ".sizeof($result2));
        foreach ($result2 as $row) {
            $index = substr($row['post_date'], 0, 7);
            switch ($row['journal_id']) {
                case  6:
                case 21: $history['purchases'][$index]['qty']   += $row['qty'];
                         $history['purchases'][$index]['usage'] += $row['qty'];
                         $history['purchases'][$index]['total'] += $row['debit_amount']; break;
                case  7: $history['purchases'][$index]['qty']   -= $row['qty'];
                         $history['purchases'][$index]['usage'] -= $row['qty'];
                         $history['purchases'][$index]['total'] -= $row['debit_amount']; break;
                case 12:
                case 19: $history['sales'][$index]['qty']       += $row['qty'];
                         $history['sales'][$index]['usage']     += $row['qty'];
                         $history['sales'][$index]['total']     += $row['credit_amount']; break;
                case 13: $history['sales'][$index]['qty']       -= $row['qty'];
                         $history['sales'][$index]['usage']     -= $row['qty'];
                         $history['sales'][$index]['total']     -= $row['debit_amount']; break;
                case 14: if ($row['gl_type'] == 'asi') { $history['sales'][$index]['usage'] -= $row['qty']; } break;
                case 16: $history['sales'][$index]['usage']     += $row['qty']; break;
            }
        }
        // calculate average usage
        $history['01purch']  = 0;
        $history['03purch']  = 0;
        $history['06purch']  = 0;
        $history['12purch']  = 0;
        $cnt = 0;
        $history['purchases']= array_values($history['purchases']);
        foreach ($history['purchases'] as $key => $value) {
            if ($cnt == 0) { $cnt++; continue; } // skip current month since we probably don't have the full months worth
            $history['12purch']               += $value['usage'];
            if ($cnt < 7) { $history['06purch'] += $value['usage']; }
            if ($cnt < 4) { $history['03purch'] += $value['usage']; }
            if ($cnt < 2) { $history['01purch'] += $value['usage']; }
            $cnt++;
        }
        $history['12purch'] = round($history['12purch'] / 12);
        $history['06purch'] = round($history['06purch'] /  6);
        $history['03purch'] = round($history['03purch'] /  3);

        $history['01sales'] = $history['02sales'] = $history['03sales'] = $history['06sales'] = $history['12sales'] = 0;
        $cnt1  = 0;
        $sales = [];
        $history['sales']   = array_values($history['sales']);
        foreach ($history['sales'] as $key => $value) {
            if ($cnt1 == 0) { $cnt1++; continue; }
            $history['12sales']               += $value['usage'];
            if ($cnt1 < 7) { $history['06sales'] += $value['usage']; }
            if ($cnt1 < 4) { $history['03sales'] += $value['usage']; }
            if ($cnt1 < 3) { $history['02sales'] += $value['usage']; }
            if ($cnt1 < 2) { $history['01sales'] += $value['usage']; }
            if ($cnt1 <= $this->months_of_data) { $sales[] = $value['usage']; }
            $cnt1++;
        }
        $history['12sales'] = round($history['12sales'] / 12);
        $history['06sales'] = round($history['06sales'] /  6);
        $history['03sales'] = round($history['03sales'] /  3);
        // find the restock levels that need adjustment
        if (getModuleCache('inventory', 'settings', 'general', 'stock_usage')
                && validateSecurity('phreebooks', 'j6_mgr', 3, false)
                && in_array($skuInfo['inventory_type'], explode(',', COG_ITEM_TYPES))) {
//          $inv = dbGetValue(BIZUNO_DB_PREFIX."inventory", ['qty_min', 'lead_time'], "sku='{$skuInfo['sku']}'");
            sort($sales);
            $months        = substr('0'.$this->months_of_data, -2);
            $idx           = ceil(count($sales) / 2);
            $median_sales  = $sales[$idx];
            $average_sales = ceil($history[$months.'sales']);
            $new_min_stock = ceil($skuInfo['lead_time'] / 30) * $average_sales;
            $high_band     = $skuInfo['qty_min'] * (1 + $this->percent_diff);
            $low_band      = $skuInfo['qty_min'] * (1 - $this->percent_diff);
            $high_avg      = $average_sales  * (1 + $this->med_avg_diff);
            $low_avg       = $average_sales  * (1 - $this->med_avg_diff);
            if ($new_min_stock > $high_band || $new_min_stock < $low_band) {
                if ($verbose) { msgAdd(sprintf($this->lang['msg_inv_qty_min'], $new_min_stock), 'caution'); }
            }
            if ($median_sales > $high_avg || $median_sales < $low_avg) {
                if ($verbose) { msgAdd(sprintf($this->lang['msg_inv_median'], $median_sales, $average_sales), 'caution'); }
            }
        }
        return $history;
    }

    /**
     * Grid structure for PO/SO history with options
     * @param type $jID
     * @return type
     */
    private function dgJ04J10($jID=10, $skuID=0)
    {
        $hide_cost= validateSecurity('phreebooks', "j6_mgr", 1, false) ? false : true;
        $stores   = sizeof(getModuleCache('bizuno', 'stores'))>1 ? false : true;
$label = 'TBD';
$icon = 'phreesoft';
$invID = 0;
        switch ($jID) {
            case  3:
                $props = ['name'=>'dgJ03','title'=>lang('open_journal_3'), 'data'=>'dataPQ'];
                $hide  = validateSecurity('phreebooks', "j3_mgr", 1, false);
                break;
            case  4:
                $props = ['name'=>'dgJ04','title'=>lang('open_journal_4'), 'data'=>'dataPO'];
                $label = lang('fill_purchase');
                $invID = 6;
                $icon  = 'purchase';
                $hide  = validateSecurity('phreebooks', "j4_mgr", 1, false);
                break;
            case  9:
                $props = ['name'=>'dgJ09','title'=>lang('open_journal_9'), 'data'=>'dataSQ'];
                $hide  = validateSecurity('phreebooks', "j3_mgr", 1, false);
                break;
            case 10:
                $props = ['name'=>'dgJ10','title'=>lang('open_journal_10'),'data'=>'dataSO'];
                $label = lang('fill_sale');
                $invID = 12;
                $icon  = 'sales';
                $hide  = validateSecurity('phreebooks', "j10_mgr", 1, false);
                break;
        }
        $data = ['id' => $props['name'],
            'attr'   => ['title'=>$props['title'], 'pagination'=>false, 'idField'=>'id'],
            'events' => ['url'=>"'".BIZUNO_AJAX."&bizRt=inventory/history/historyRows&jID=$jID&skuID=$skuID'"],
            'columns'=> ['id'   => ['attr'=>['hidden'=>true]],
                'action'        => ['order'=>1,'label'=>lang('action'),'attr'=>['hidden'=>$hide_cost?true:false],
                    'events'    => ['formatter'=>"function(value,row,index) { return {$props['name']}Formatter(value,row,index); }"],
                    'actions'   => [
                        'edit'  => ['order'=>20,'icon'=>'edit',  'label'=>lang('edit'),          'hidden'=>$hide>0?false:true,'events'=>['onClick'=>"winHref(bizunoHome+'&bizRt=phreebooks/main/manager&rID=idTBD');"]],
                        'toggle'=> ['order'=>40,'icon'=>'toggle','label'=>lang('toggle_status'), 'hidden'=>$hide>2?false:true,'events'=>['onClick'=>"jsonAction('phreebooks/main/toggleWaiting&jID=$jID&dgID={$props['name']}', idTBD);"]],
                        'dates' => ['order'=>50,'icon'=>'date',  'label'=>lang('delivery_dates'),'hidden'=>$hide>2?false:true,'events'=>['onClick'=>"windowEdit('phreebooks/main/deliveryDates&rID=idTBD', 'winDelDates', '".lang('delivery_dates')."', 500, 400);"]],
                        'fill'  => ['order'=>80,'icon'=>$icon,   'label'=>$label,                'hidden'=>$hide>2?false:true,'events'=>['onClick'=>"winHref(bizunoHome+'&bizRt=phreebooks/main/manager&rID=idTBD&jID=$invID&bizAction=inv');"]]]],
                'invoice_num'   => ['order'=>20,'label'=>lang('journal_main_invoice_num', $jID),'attr'=>['resizable'=>true]],
                'primary_name_b'=> ['order'=>30,'label'=>lang('address_book_primary_name'),'attr'=>['sortable'=>true,'resizable'=>true]],
                'purch_order_id'=> ['order'=>40,'label'=>lang('journal_main_purch_order_id', $jID),'attr'=>['resizable'=>true]],
                'store_id'      => ['order'=>50,'label'=>lang('contacts_short_name_b'),   'attr'=>['resizable'=>true,'hidden'=>$stores]],
                'rep_id'        => ['order'=>60,'label'=>lang('contacts_rep_id_c'),       'attr'=>['resizable'=>true,'align'=>'center']],
                'post_date'     => ['order'=>70,'label'=>lang('post_date'),               'attr'=>['resizable'=>true,'sortable'=>true,'align'=>'center']],
                'qty'           => ['order'=>80,'label'=>lang('balance'),                 'attr'=>['resizable'=>true,'align'=>'center']],
                'date_1'        => ['order'=>90,'label'=>jsLang('journal_item_date_1',10),'attr'=>['resizable'=>true,'sortable'=>true,'align'=>'center'],
                    'events'=>['styler'=>"function(value,row,index) { if (row.waiting==1) { return {style:'background-color:yellowgreen'}; } }"]]]];
        if (in_array($jID, [3,9])) { // remove some action items
            unset($data['columns']['action']['actions']['toggle'],$data['columns']['action']['actions']['dates'],$data['columns']['action']['actions']['fill']);
        }
        return $data;
    }

    private function dgJ06J12($jID=6, $skuID=0, $sku='')
    {
        $rows = clean('rows', ['format'=>'integer','default'=>getModuleCache('bizuno', 'settings', 'general', 'max_rows')], 'post');
        $page = clean('page', ['format'=>'integer','default'=>1], 'post');
        $sort = clean('sort', ['format'=>'text',   'default'=>'post_date'],'post');
        $order= clean('order',['format'=>'text',   'default'=>'desc'],     'post');
        $gID  = explode(":", getDefaultFormID($jID)); // used for printing from popup
        $name = "dgMvmt$jID";
        switch ($jID) {
            case  6: $jPmt = 20; break;
            case 12:
            default: $jPmt = 18; break;
        }
        $data = ['id'=>$name, 'strict'=>true, 'rows'=>$rows, 'page'=>$page, 'title'=>sprintf(lang('tbd_history'), lang('journal_main_journal_id', $jID)),
            'attr'   => ['idField'=>'id','url'=>BIZUNO_AJAX."&bizRt=inventory/history/movementRows&jID=$jID&skuID=$skuID"],
            'source' => [
                'tables' => [
                    'journal_main'=>['table'=>BIZUNO_DB_PREFIX.'journal_main'],
                    'journal_item'=>['table'=>BIZUNO_DB_PREFIX.'journal_item','join'=>'join','links'=>BIZUNO_DB_PREFIX."journal_main.id=".BIZUNO_DB_PREFIX."journal_item.ref_id"]],
                'filters' => [
                    'jID'  => ['order'=>99, 'hidden'=>true, 'sql'=>BIZUNO_DB_PREFIX."journal_main.journal_id=$jID"],
                    'rID'  => ['order'=>99, 'hidden'=>true, 'sql'=>BIZUNO_DB_PREFIX."journal_item.sku='$sku'"]],
                'sort' => ['s0'=>['order'=>10, 'field'=>("$sort $order")]]],
            'columns' => [
                'id'        => ['order'=>0, 'field'=>BIZUNO_DB_PREFIX.'journal_main.id',        'attr'=>['hidden'=>true]],
                'closed'    => ['order'=>0, 'field'=>BIZUNO_DB_PREFIX.'journal_main.closed',    'attr'=>['hidden'=>true]],
                'journal_id'=> ['order'=>0, 'field'=>BIZUNO_DB_PREFIX.'journal_main.journal_id','attr'=>['hidden'=>true]],
                'bal_due'   => ['order'=>0, 'field'=>BIZUNO_DB_PREFIX.'journal_main.id','process'=>'paymentRcv','attr'=>['hidden'=>true]],
                'action'    => ['order'=>1, 'label'=>lang('action'),'events'=>['formatter'=>"function(value,row,index){ return ".$name."Formatter(value,row,index); }"],
                    'actions'=> [
                        'edit'       => ['order'=>20,'icon'=>'edit',    'label'=>lang('edit'),
                            'events' => ['onClick' => "winHref(bizunoHome+'&bizRt=phreebooks/main/manager&rID=idTBD');"]],
                        'print'      => ['order'=>40,'icon'=>'print',   'label'=>lang('print'),
                            'events' => ['onClick'=>"var idx=jqBiz('#$name').datagrid('getRowIndex', idTBD); var jID=jqBiz('#$name').datagrid('getRows')[idx].journal_id; ('fitColumns', true); winOpen('phreeformOpen', 'phreeform/render/open&group={$gID[0]}:j'+jID+'&date=a&xfld=journal_main.id&xcr=equal&xmin=idTBD');"]],
                                    ]],
                'invoice_num'   => ['order'=>10, 'field'=>BIZUNO_DB_PREFIX.'journal_main.invoice_num','label'=>pullTableLabel('journal_main', 'invoice_num', $jID),
                    'attr' => ['sortable'=>true, 'resizable'=>true]],
                'primary_name_b'=> ['order'=>20, 'field'=>BIZUNO_DB_PREFIX.'journal_main.primary_name_b','label'=>pullTableLabel('address_book', 'primary_name', '_b'),
                    'attr'  => ['sortable'=>true, 'resizable'=>true]],
                'purch_order_id'=> ['order'=>30, 'field'=>BIZUNO_DB_PREFIX.'journal_main.purch_order_id','label'=>pullTableLabel('journal_main', 'purch_order_id', $jID),
                    'attr'  => ['sortable'=>true, 'resizable'=>true]],
                'store_id'      => ['order'=>40, 'field'=>BIZUNO_DB_PREFIX.'journal_main.store_id','label'=>pullTableLabel('journal_main', 'store_id', $jID),'format'=>'storeID',
                    'attr'  => ['sortable'=>false, 'resizable'=>true]],
                'post_date'     => ['order'=>50, 'field' => BIZUNO_DB_PREFIX.'journal_main.post_date', 'format'=>'date','label' => pullTableLabel('journal_main', 'post_date', $jID),
                    'attr' => ['align'=>'center', 'sortable'=>true, 'resizable'=>true]],
                'qty'           => ['order'=>60, 'field' => BIZUNO_DB_PREFIX.'journal_item.qty', 'format'=>'integer','label' => lang('quantity'),
                    'attr' => ['align'=>'center', 'sortable'=>true, 'resizable'=>true]],
                ]];
        return $data;
    }

    /**
     * Lists open Work Orders for a give SKU
     * @param integer $skuID - id field from the inventory db
     */
    private function dgWO($skuID=0)
    {
        $rows  = clean('rows', ['format'=>'integer','default'=>getModuleCache('bizuno', 'settings', 'general', 'max_rows')], 'post');
        $page  = clean('page', ['format'=>'integer','default'=>1],            'post');
        $sort  = clean('sort', ['format'=>'text',   'default'=>'create_date'],'post');
        $order = clean('order',['format'=>'text',   'default'=>'desc'],       'post');
        $hide  = 1; //getUserCache('security', 'srvBldJrnl') ? false : true;
//      $stores= sizeof(getModuleCache('bizuno', 'stores')) > 1 ? false : true;
        $data  = ['id'=>'dgWO', 'strict'=>true, 'rows'=>$rows, 'page'=>$page, 'title'=>'Open Work Orders',
            'attr'    => ['idField'=>'id','url'=>BIZUNO_AJAX."&bizRt=inventory/history/buildRows&skuID=$skuID"],
            'source'  => [
                'tables' => ['srvBuilder_journal'=>['table'=>BIZUNO_DB_PREFIX.'srvBuilder_journal']],
                'filters' => [
                    'closed'=> ['order'=>99, 'hidden'=>true, 'sql'=>"closed='0'"],
                    'rID'   => ['order'=>99, 'hidden'=>true, 'sql'=>"sku_id='$skuID'"]],
                'sort' => ['s0'=>['order'=>10, 'field'=>("$sort $order")]]],
            'columns' => [
                'id'         => ['order'=>1,'attr'=>['hidden'=>true]],
                'started'    => ['order'=>1,'attr'=>['hidden'=>true]],
                'action'     => ['order'=>5,'label'=>lang('action'),
                    'events' => ['formatter'=>"function(value,row,index) { return dgWOFormatter(value,row,index); }"],
                    'actions'=> ['print'=>['order'=>20,'icon'=>'print','label'=>lang('print'),'hidden'=>$hide>0?false:true,'events'=>['onClick'=>"winOpen('phreeformOpen', 'phreeform/render/open&group=mfg:wo&date=a&xfld=srvBuilder_journal.id&xcr=equal&xmin=idTBD');"]]]],
                'sb_ref'     => ['order'=>10,'field'=>'sb_ref','label'=>'WO Reference',                      'attr'=>['align'=>'center','sortable'=>true,'resizable'=>true],
                    'events' => ['styler'=>"function(value,row,index) { if (row.started=='1') { return {style:'background-color:lightgreen'}; } }"]],
                'store_id'   => ['order'=>20,'field'=>'store_id','label'=>lang('store_id'),                  'attr'=>['align'=>'center','sortable'=>true,'resizable'=>true]],
                'create_date'=> ['order'=>30,'field'=>'create_date','format'=>'date','label'=>'Post Date',   'attr'=>['align'=>'center','sortable'=>true,'resizable'=>true]],
                'due_date'   => ['order'=>40,'label'=>lang('due_date'),'field'=>'due_date','format'=>'date', 'attr'=>['align'=>'center','sortable'=>true,'resizable'=>true]],
                'qty'        => ['order'=>50,'field'=>'qty',   'format'=>'integer','label'=>lang('quantity'),'attr'=>['align'=>'center','sortable'=>true,'resizable'=>true]],
                'notes'      => ['order'=>60,'field'=>'notes','label'=>lang('notes'),'attr'=>['sortable'=>true,'resizable'=>true]]],
            'footnotes'=> ['status'=>lang('status').': <span style="background-color:lightgreen">'.'Work Order Started'.'</span>']];
        return $data;
    }

    private function dgJ06J12Avg($jID=12)
    {
        if ($jID==6) {
            $props = ['name'=>'dgJ06','title'=>sprintf(lang('tbd_history'), lang('journal_main_journal_id', '6')), 'data'=>'dataJ6'];
            $label = jsLang('cost');
        } else {
            $props = ['name'=>'dgJ12','title'=>sprintf(lang('tbd_history'), lang('journal_main_journal_id', '12')),'data'=>'dataJ12'];
            $label = jsLang('sales');
        }
        return ['id' => $props['name'],
            'attr'   => ['title'=>$props['title'], 'pagination'=>false],
            'events' => ['data' =>$props['data']],
            'columns'=> [
                'year' => ['order'=>20,'label'=>lang('year'), 'attr'=>['align'=>'center','resizable'=>true]],
                'month'=> ['order'=>30,'label'=>lang('month'),'attr'=>['align'=>'center','resizable'=>true]],
                'qty'  => ['order'=>40,'label'=>lang('qty'),  'attr'=>['align'=>'center','resizable'=>true]],
                'total'=> ['order'=>50,'label'=>$label,       'attr'=>['align'=>'right', 'resizable'=>true],'events'=>['formatter'=>"function(value) { return formatCurrency(value); }"]]]];
    }
}
